-----------------------------------------------------------------------------
-- |
-- Copyright   :  (C) 2011-2015,2018 Edward Kmett
-- License     :  BSD-style (see the file LICENSE)
--
-- Maintainer  :  Edward Kmett <ekmett@gmail.com>
-- Stability   :  provisional
-- Portability :  polykinds
--
-- Re-exports from the `base-orphans` and `transformers-compat` packages.
----------------------------------------------------------------------------
module Data.Traversable.Instances where

import Control.Monad.Trans.Instances ()
import Data.Orphans ()
